@@ -94,8 +94,8 @@ class CommandLineParser(object):
                           "type": str},
                     'V': {"short": '-V',
                           "long": '--version',
-                          "help": 'Hadoop protocol version (default:8)',
-                          "default": 7,
+                          "help": 'Hadoop protocol version (default:9)',
+                          "default": 9,
                           "type": float},
                     'p': {"short": '-p',
                           "long": '--port',
@@ -257,7 +257,7 @@ class CommandLineParser(object):
             config = json.loads(open(os.path.join(os.path.expanduser('~'), '.snakebiterc')).read())
             self.args.namenode = config['namenode']
             self.args.port = config['port']
-            self.args.version = config.get('version', 7)
+            self.args.version = config.get('version', 9)
         elif os.environ.get('HADOOP_HOME'):
             hdfs_conf = os.path.join(os.environ['HADOOP_HOME'], 'conf', 'core-site.xml')
             self._read_hadoop_config(hdfs_conf, config_file)
@@ -277,7 +277,7 @@ class CommandLineParser(object):
             for hdfs_conf in try_paths:
                 print " - %s" % hdfs_conf
             print "\nYou can manually create ~/.snakebiterc with the following content:"
-            print '{"namenode": "ip/hostname", "port": 54310, "version": 7}'
+            print '{"namenode": "ip/hostname", "port": 54310, "version": 9}'
             sys.exit(1)
 
     def _read_hadoop_config(self, hdfs_conf, config_file):
@@ -540,7 +540,15 @@ class CommandLineParser(object):
     def cat(self):
         for file_to_read in self.client.cat(self.args.dir, check_crc=self.args.checkcrc):
             for load in file_to_read:
-                print load
+                sys.stdout.write(load)
+
+    @command(args="path dst", descr="copy local file reference to destination", req_args=['dir [dirs]', 'arg'])
+    def copyFromLocal(self):
+        src = self.args.dir
+        dst = self.args.single_arg
+        result = self.client.copyFromLocal(src, dst)
+        for line in format_results(result, json_output=self.args.json):
+            print line
 
     @command(args="[paths] dst", descr="copy paths to local file system destination", allowed_opts=['checkcrc'], req_args=['dir [dirs]', 'arg'])
     def copyToLocal(self):
@@ -550,6 +558,14 @@ class CommandLineParser(object):
         for line in format_results(result, json_output=self.args.json):
             print line
 
+    @command(args="[paths] dst", descr="copy files from source to destination", allowed_opts=['checkcrc'], req_args=['dir [dirs]', 'arg'])
+    def cp(self):
+        paths = self.args.dir
+        dst = self.args.single_arg
+        result = self.client.cp(paths, dst, checkcrc=self.args.checkcrc)
+        for line in format_results(result, json_output=self.args.json):
+            print line
+
     @command(args="file dst", descr="copy files to local file system destination", allowed_opts=['checkcrc'], req_args=['dir [dirs]', 'arg'])
     def get(self):
         paths = self.args.dir
@@ -566,6 +582,14 @@ class CommandLineParser(object):
         for line in format_results(result, json_output=self.args.json):
             print line
 
+    # @command(args="[paths] dst", descr="copy sources from local file system to destination", req_args=['dir [dirs]', 'arg'])
+    # def put(self):
+    #     paths = self.args.dir
+    #     dst = self.args.single_arg
+    #     result = self.client.put(paths, dst)
+    #     for line in format_results(result, json_output=self.args.json):
+    #         print line
+
     @command(args="path", descr="display last kilobyte of the file to stdout", allowed_opts=['f'], req_args=['arg'])
     def tail(self):
         path = self.args.single_arg
